@@ -2,7 +2,8 @@
 
 #
 # dbcolstats.pm
-# Copyright (C) 1991-2018 by John Heidemann <johnh@isi.edu>
+# Copyright (C) 1991-2007 by John Heidemann <johnh@isi.edu>
+# $Id$
 #
 # This program is distributed under terms of the GNU general
 # public license, version 2.  See the file COPYING
@@ -73,14 +74,14 @@ Compute median value.  (Will sort data if necessary.)
 =item B<-q N> or B<--quantile N>
 
 Compute quantile (quartile when N is 4),
-or an arbitrary quantile for other values of N,
+or an artbitrary quantile for other values of N,
 where the scores that are 1 Nth of the way across the population.
 
 =item B<--sample>
 
 Compute I<sample> population statistics
-(e.g., the sample standard deviation),
-assuming I<n-1> degrees of freedom.
+(e.g., the sample standard devation),
+assuming I<n-1> degress of freedom.
 
 =item B<--nosample>
 
@@ -92,13 +93,6 @@ Compute I<whole> population statistics
 Assume data is already sorted.
 With one -S, we check and confirm this precondition.
 When repeated, we skip the check.
-(This flag is ignored if quartiles are not requested.)
-
-=item B<--parallelism=N> or C<-j N>
-
-Allow sorting to happen in parallel.
-Defaults on.
-(Only relevant if using non-pre-sorted data with quantiles.)
 
 =item B<-F> or B<--fs> or B<--fieldseparator> S
 
@@ -113,17 +107,6 @@ Also uses environment variable TMPDIR, if -T is
 not specified.
 Default is /tmp.
 
-=item B<-k KeyField>
-
-Do multi-stats, grouped by each key.
-Assumes keys are sorted.  (Use dbmultistats to guarantee sorting order.)
-
-=item B<--output-on-no-input>
-
-Enables null output (all fields are "-", n is 0)
-if we get input with a schema but no records.
-Without this option, just output the schema but no rows.
-Default: no output if no input.
 
 =back
 
@@ -240,8 +223,10 @@ See the test cases F<dbcolstats_extrema> for examples.
 
 use strict;
 use Pod::Usage;
-use Carp;
+use threads;
+use threads::shared;
 
+use Fsdb::BoundedQueue;
 use Fsdb::IO::Reader;
 use Fsdb::IO::Writer;
 use Fsdb::Filter;
@@ -255,11 +240,11 @@ use Fsdb::Support::NamedTmpfile;
 
     $filter = new Fsdb::Filter::dbcolstats(@arguments);
 
-Create a new dbcolstats object, taking command-line arguments.
+Create a new dbcolstats object, taking command-line arugments.
 
 =cut
 
-sub new($@) {
+sub new ($@) {
     my $class = shift @_;
     my $self = $class->SUPER::new(@_);
     bless $self, $class;
@@ -278,7 +263,7 @@ Internal: set up defaults.
 
 =cut
 
-sub set_defaults($) {
+sub set_defaults ($) {
     my($self) = @_;
     $self->SUPER::set_defaults();
     $self->{_target_column} = undef;
@@ -290,10 +275,6 @@ sub set_defaults($) {
     $self->{_pre_sorted} = 0;
     $self->{_include_non_numeric} = undef;
     $self->{_fscode} = undef;
-    $self->{_max_parallelism} = undef;
-    $self->{_key_column} = undef;
-    $self->{_output_on_no_input} = undef;
-    $self->set_default_tmpdir;
 }
 
 =head2 parse_options
@@ -304,7 +285,7 @@ Internal: parse command-line arguments.
 
 =cut
 
-sub parse_options($@) {
+sub parse_options ($@) {
     my $self = shift @_;
 
     my(@argv) = @_;
@@ -318,19 +299,15 @@ sub parse_options($@) {
 	'c|confidence=f' => \$self->{_confidence_fraction},
 	'd|debug+' => \$self->{_debug},
 	'f|format=s' => \$self->{_format},
-	'F|fs|cs|fieldseparator|columnseparator=s' => \$self->{_fscode},
+	'F|fs|cs|fieldseperator|columnseperator=s' => \$self->{_fscode},
 	'i|input=s' => sub { $self->parse_io_option('input', @_); },
-	'j|parallelism=i' => \$self->{_max_parallelism},
-	'k|key=s' => \$self->{_key_column},
 	'log!' => \$self->{_logprog},
 	'm|median!' =>  \$self->{_median},
 	'o|output=s' => sub { $self->parse_io_option('output', @_); },
-        'output-on-no-input!' => \$self->{_output_on_no_input},
 	'q|quantile=i' => \$self->{_quantile},
 	's|sample!' =>  \$self->{_sample},
 	'S|pre-sorted+' => \$self->{_pre_sorted},
 	'T|tmpdir|tempdir=s' => \$self->{_tmpdir},
-	'saveoutput=s' => \$self->{_save_output},
 	) or pod2usage(2);
     $self->parse_target_column(\@argv);
 }
@@ -344,7 +321,7 @@ Internal: setup, parse headers.
 =cut
 
 
-sub setup($) {
+sub setup ($) {
     my($self) = @_;
 
     pod2usage(2) if (!defined($self->{_target_column}));
@@ -353,14 +330,8 @@ sub setup($) {
     $self->finish_io_option('input', -comment_handler => $self->create_delay_comments_sub);
     print STDERR "dbcolstats: post-input setup\n" if ($self->{_debug} > 2);
     $self->{_target_coli} = $self->{_in}->col_to_i($self->{_target_column});
-    croak($self->{_prog} . ": target column " . $self->{_target_column} . " is not in input stream.\n")
+    croak $self->{_prog} . ": target column " . $self->{_target_column} . " is not in input stream.\n"
 	if (!defined($self->{_target_coli}));
-    $self->{_key_coli} = undef;
-    if (defined($self->{_key_column})) {
-        $self->{_key_coli} = $self->{_in}->col_to_i($self->{_key_column});
-        croak($self->{_prog} . ": key column " . $self->{_key_column} . " is not in input stream.\n")
-	    if (!defined($self->{_key_coli}));
-    };
     my $read_fastpath_sub = $self->{_in}->fastpath_sub();
     $self->{_read_fastpath_sub} = $read_fastpath_sub;
 
@@ -372,7 +343,6 @@ sub setup($) {
 	    push(@headers, "q$_");
 	};
     };
-    unshift(@headers, $self->{_key_column}) if (defined($self->{_key_column}));
     print STDERR "dbcolstats: pre-output setup\n" if ($self->{_debug} > 2);
     my @output_options = (-cols => \@headers);
     unshift (@output_options, -fscode => $self->{_fscode})
@@ -381,23 +351,18 @@ sub setup($) {
     print STDERR "dbcolstats: post-output setup\n" if ($self->{_debug} > 2);
 
     if ($self->{_quantile} || $self->{_median}) {
-        croak($self->{_prog} . ": cannot currently do median or quantile with a key column\n")
-	    if (defined($self->{_key_column}));
 	$self->{_save_out_filename} = Fsdb::Support::NamedTmpfile::alloc($self->{_tmpdir});
 	# sorting needed?
 	my $save_out;
 	my(@writer_args) = (-cols => [qw(data)]);
 	print STDERR "dbcolstats: pre-saveoutput setup\n" if ($self->{_debug} > 2);
 	if (!$self->{_pre_sorted}) {
-	    my $sorter_fred;
-	    my(@dbsort_args) = qw(-n data);
-	    push (@dbsort_args, '--parallelism', $self->{_max_parallelism})
-		if (defined($self->{_max_parallelism}));
+	    my $sorter_thread;
 	    print STDERR "dbcolstats: doing sorter thread\n" if ($self->{_debug} > 2);
-	    ($save_out, $sorter_fred) = dbpipeline_sink(\@writer_args,
+	    ($save_out, $sorter_thread) = dbpipeline_sink(\@writer_args,
 			'--output' => $self->{_save_out_filename},
-			dbsort(@dbsort_args));
-	    $self->{_sorter_fred} = $sorter_fred;
+			dbsort(qw(-n data)));
+	    $self->{_sorter_thread} = $sorter_thread;
 	} else {
 	    # no, just write it ourselves
 	    $save_out = new Fsdb::IO::Writer('-file' => $self->{_save_out_filename}, @writer_args);
@@ -418,7 +383,7 @@ Internal: Round up to the next integer.
 
 =cut
 
-sub _round_up($) {
+sub _round_up {
     my($x) = @_;
     my($xi) = int($x);
     return ($x > $xi) ? $xi+1 : $xi;
@@ -434,7 +399,7 @@ We assume the saved output is closed before we enter.
 
 =cut
 
-sub _compute_quantile($$$) {
+sub _compute_quantile {
     my ($self, $n, $mean) = @_;
 
     return if (!($self->{_quantile} || $self->{_median}));
@@ -450,7 +415,7 @@ sub _compute_quantile($$$) {
     };
 
     my $save_in = new Fsdb::IO::Reader(-file => $self->{_save_out_filename});
-    $save_in->error && croak($self->{_prog} . ": re-read error " . $save_in->error);
+    $save_in->error && die $self->{_prog} . ": re-read error " . $save_in->error;
 
     # To handle the ugly case of having more ntiles than
     # data, we detect it and replicate the data until we have more
@@ -470,14 +435,14 @@ sub _compute_quantile($$$) {
     for ($i = 1; $#q+1 < $effective_quantile; $i++) {
 	if (--$replicates_left <= 0) {
 	    my $fref = $save_in->read_rowobj;
-	    croak("internal error re-reading data\n") if (ref($fref) ne 'ARRAY');
+	    die "internal error re-reading data\n" if (ref($fref) ne 'ARRAY');
 	    $x = $fref->[0];
 	    $replicates_left = $replicate_data;
 	    # Verify sorted order (in case the user lied to us
 	    # about pre-sorting).
 	    if (defined($last_x) && $x < $last_x) { 
 		my($info) = ($self->{_pre_sorted} ? " (internal error in dbsort)" : " (user specified -S for pre-sorted data but it is unsorted)");
-		croak($self->{_prog} . ": cannot process data that is out of order between $last_x and $x $info.\n");
+		die $self->{_prog} . ": cannot process data that is out of order between $last_x and $x $info.\n";
 	    };
 	    $last_x = $x;
 	};
@@ -490,31 +455,26 @@ sub _compute_quantile($$$) {
 };
 
 
-=head2 run_one_key
+=head2 run
 
-    $filter->run_one_key();
+    $filter->run();
 
-Internal: run over each row, for a given key.
+Internal: run over each rows.
 
 =cut
-sub run_one_key($) {
+sub run ($) {
     my($self) = @_;
 
     print STDERR "dbcolstats: starting run\n" if ($self->{_debug} > 2);
 
     # xxx: should eval all this to factor out constants from runtime
     my($xf) = $self->{_target_coli};
-    my($key_column) = $self->{_key_column};
 
     my($n) = 0;
     my($sx) = 0;
     my($sxx) = 0;
     my $min;
     my $max;
-    my $key;
-    my $last_key = $self->{_holdover_key};
-    my $holdover_data = $self->{_holdover_data};
-    $self->{_holdover_key} = $self->{_holdover_data} = undef;
 
     my $fref;
     my $x;
@@ -524,31 +484,9 @@ sub run_one_key($) {
 	my $read_fastpath_sub = $self->{_read_fastpath_sub};
 
 	my $code = q'
-	    while (1) {
-		if (defined($holdover_data)) {
-		    $x = $holdover_data;   # and key was set earlier
-		    $holdover_data = undef;
-		} else {
-		    $fref = &{$read_fastpath_sub}();
-		    last if (!defined($fref));
-		    $x = $fref->[' . $xf . q'];
-		';
-	if (defined($self->{_key_column})) {
-	    $code .= q'
-		    $key = $fref->[' . $self->{_key_coli} . '];
-		    if (!defined($last_key)) {
-			$last_key = $key;
-		    } elsif ($key ne $last_key) {
-			$self->{_holdover_key} = $key;
-			$self->{_holdover_data} = $x;
-			last;
-		    };
-	    ';
-	};
-	$code .= q'	       		    
-		};
+	    while ($fref = &{$read_fastpath_sub}()) {
+		$x = $fref->[' . $xf . q'];
 	';
-
 	$code .= 'next if ($x !~ /' . $is_numeric_regexp . "/);\n"
 	    if (!$self->{_include_non_numeric});
 	$code .= q'
@@ -581,7 +519,7 @@ sub run_one_key($) {
 	print STDERR "dbcolstats: eval'ing code\n" if ($self->{_debug});
 	print $code if ($self->{_debug});
 	eval $code;
-	$@ and croak($self->{_prog} . ": internal error in eval.: $@\n");
+	$@ and die $self->{_prog} . ": internal error in eval.: $@\n";
 
 	# clean up
        	if ($self->{_quantile} || $self->{_median}) {
@@ -594,11 +532,11 @@ sub run_one_key($) {
     #
     # Make sure we cleaned up before we do any computation.
     #
-    if (defined($self->{_sorter_fred})) {
+    if (defined($self->{_sorter_thread})) {
 	# let sorting finish
 	print STDERR "dbcolstats: join on sorter thread\n" if ($self->{_debug} > 2);
-	$self->{_sorter_fred}->join();
-	$self->{_sorter_fred} = undef;
+	$self->{_sorter_thread}->join;
+	$self->{_sorter_thread} = undef;
 	print STDERR "dbcolstats: post join on sorter thread\n" if ($self->{_debug} > 2);
     };
 
@@ -616,13 +554,8 @@ sub run_one_key($) {
     if ($n == 0) {
 	$stddev = "-";
     } else {
-        my $sxx_less_mean_sq = $sxx - $n * $mean * $mean;
-        if ($sxx_less_mean_sq < 0) {
-            $stddev = 0;
-        } else {
-                $stddev = ($n <= 1 || $max == $min) ? 0 : 
-                    sqrt($sxx_less_mean_sq / ($n - ($self->{_sample} ? 1 : 0)));
-        };
+        $stddev = ($n <= 1 || $max == $min) ? 0 : 
+	    sqrt(($sxx - $n * $mean * $mean) / ($n - ($self->{_sample} ? 1 : 0)));
     };
     my $pct_rsd;
     if ($stddev eq '-' || $mean eq '-' || $mean == 0) {
@@ -677,34 +610,13 @@ sub run_one_key($) {
 	    $out_hash{"q$_"} = $q_aref->[$_];
 	};
     };
-    if (defined($key_column)) {
-        $out_hash{$key_column} = $last_key;
-    };
 
-    if ($n > 0 || ($n == 0 && $self->{_output_on_no_input})) {
-        $self->{_out}->write_row_from_href(\%out_hash);
-    };
-}
-
-=head2 run
-
-    $filter->run();
-
-Internal: run over each row, for one or many keys.
-
-=cut
-sub run($) {
-    my($self) = @_;
-    $self->{_holdover_key} = $self->{_holdove_data} = undef;
-    for (;;) {
-	$self->run_one_key();
-	last if (!defined($self->{_holdover_key}));
-    };
+    $self->{_out}->write_row_from_href(\%out_hash);
 }
 
 =head1 AUTHOR and COPYRIGHT
 
-Copyright (C) 1991-2018 by John Heidemann <johnh@isi.edu>
+Copyright (C) 1991-2007 by John Heidemann <johnh@isi.edu>
 
 This program is distributed under terms of the GNU general
 public license, version 2.  See the file COPYING
